#ifndef bqtXYZhh
#define bqtXYZhh

#include <cstdio>
#include "floatdefs.hh"
#include "matrix.hh"

class XYZ
{
public:
    FLOAT x,y,z;
    XYZ() : x(0),y(0),z(0) { }
    XYZ(const XYZ &p) : x(p.x), y(p.y), z(p.z) { }
    XYZ(FLOAT_ARG X,FLOAT_ARG Y,FLOAT_ARG Z) : x(X),y(Y),z(Z) { }
    void Set(FLOAT_ARG X,FLOAT_ARG Y,FLOAT_ARG Z) { x=X; y=Y; z=Z; }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(FLOAT_ARG d) { x/=d; y/=d; z/=d; return *this; }
    const XYZ &operator *=(FLOAT_ARG d) { x*=d; y*=d; z*=d; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator /(FLOAT_ARG d) const { return XYZ(x/d, y/d, z/d); }
    const XYZ operator *(FLOAT_ARG d) const { return XYZ(x*d, y*d, z*d); }
    const XYZ operator -() const { return XYZ(-x, -y, -z); }
    XYZ &operator= (const XYZ &p) { x=p.x; y=p.y; z=p.z; return *this; }
    
    const FLOAT dist(const XYZ &p) const;
    const FLOAT squared() const { return x*x + y*y + z*z; }
    const FLOAT size() const { return SQRT(squared()); }
    
    void Normalize() { *this /= size(); }
    const XYZ Normalized() const { return *this / size(); }
    
    void Resize(FLOAT_ARG newlen) { *this *= (newlen / size()); }
    const XYZ Resized(FLOAT_ARG newlen) const { return (*this) * (newlen / size()); }
    
    void Rotate(const class Matrix &m);
    const XYZ RotatedBy(const class Matrix &m) const;
    
    void RotateAround(const XYZ &axis, FLOAT_ARG angle);
    const XYZ RotatedAround(const XYZ &axis, FLOAT_ARG angle) const;
    
    void MirrorAround(const XYZ &axis);
    const XYZ MirroredAround(const XYZ &axis) const;
    const XYZ CrossProduct(const XYZ &p) const
    {
        return XYZ(y*p.z - z*p.y,
                   z*p.x - x*p.z,
                   x*p.y - y*p.x);
    }
    const FLOAT DotProduct(const XYZ &p) const { return x*p.x + y*p.y + z*p.z; }

    void print() const
    {
        std::printf("<%.1"PF", %.1"PF", %.1"PF"> (len %.1"PF")", x,y,z, size());
    }
};

extern XYZ avg(const XYZ &a, const XYZ &b);
inline const XYZ Normalized(const XYZ &p) { return p.Normalized(); }

#define DEG2RAD(r) ((r) * M_PI / 180.0)
#define DISTORTION DEG2RAD(0.5)

#endif
